""" gather functions necessary to build an index """

import re
from typing import Optional, Tuple, Union

import faiss
from autofaiss.datasets.readers.local_iterators import read_embeddings_local, read_shapes_local
from autofaiss.datasets.readers.remote_iterators import read_embeddings_remote, read_filenames
from autofaiss.external.optimize import (
    get_optimal_batch_size,
    get_optimal_index_keys_v2,
    get_optimal_train_size,
    set_search_hyperparameters,
)
from autofaiss.indices.index_factory import index_factory
from autofaiss.utils.cast import cast_bytes_to_memory_string, to_faiss_metric_type, to_readable_time
from autofaiss.utils.decorators import Timeit


def estimate_memory_required_for_index_creation(
    nb_vectors: int, vec_dim: int, index_key: Optional[str] = None, max_index_memory_usage: Optional[str] = None
) -> Tuple[int, str]:
    """
    Estimates the RAM necessary to create the index
    The value returned is in Bytes
    """

    if index_key is None:
        if max_index_memory_usage is not None:
            index_key = get_optimal_index_keys_v2(nb_vectors, vec_dim, max_index_memory_usage)[0]
        else:
            raise ValueError("you should give max_index_memory_usage value if no index_key is given")

    # Compute the smallest number of vectors required to train the index given
    # the maximal memory constraint
    nb_vectors_train = get_optimal_train_size(nb_vectors, index_key, "1K", vec_dim)

    return 4 * nb_vectors_train * vec_dim, index_key


def get_estimated_download_time_infos(
    embeddings_hdfs_path: str, bandwidth_gbytes_per_sec: float = 1.0, indent: int = 0
) -> Tuple[str, Tuple[int, int]]:
    """
    Gives a general approximation of the download time (and preprocessing time) of embeddings
    """
    nb_vectors_approx, vec_dim = get_nb_vectors_approx_and_dim_from_hdfs(embeddings_hdfs_path)

    size = 4 * nb_vectors_approx * vec_dim

    download = 1.1 * size / (bandwidth_gbytes_per_sec * 1024 ** 3)  # seconds
    preprocess = 1.6 * download  # seconds

    infos = (
        f"-> Download: {to_readable_time(download, rounding=True)}\n"
        f"-> Preprocess: {to_readable_time(preprocess, rounding=True)}\n"
        f"Total: {to_readable_time(download + preprocess, rounding=True)}"
        " (< 1 minute if files are already cached)"
    )

    tab = "\t" * indent
    infos = tab + infos.replace("\n", "\n" + tab)

    return infos, (nb_vectors_approx, vec_dim)


def get_estimated_construction_time_infos(nb_vectors: int, vec_dim: int, indent: int = 0) -> str:
    """
    Gives a general approximation of the construction time of the index
    """

    size = 4 * nb_vectors * vec_dim

    train = 1000  # seconds, depends on the number of points for training
    add = 450 * size / (150 * 1024 ** 3)  # seconds, Linear approx (450s for 150GB in classic conditions)

    infos = (
        f"-> Train: {to_readable_time(train, rounding=True)}\n"
        f"-> Add: {to_readable_time(add, rounding=True)}\n"
        f"Total: {to_readable_time(train + add, rounding=True)}"
    )
    tab = "\t" * indent
    infos = tab + infos.replace("\n", "\n" + tab)

    return infos


def get_nb_vectors_approx_and_dim_from_hdfs(parquet_embeddings_path: str) -> Tuple[int, int]:
    """legacy function to give the dimensions of a parquet file
    Still useful for tests"""

    # Get information for one partition
    avg_batch_length, vec_dim = next(read_embeddings_remote(parquet_embeddings_path, verbose=False)).shape

    # Count the number of files
    nb_files = len(read_filenames(parquet_embeddings_path))

    nb_vectors_approx = nb_files * avg_batch_length

    return nb_vectors_approx, vec_dim


def get_nb_vectors_and_dim(embeddings_path: str) -> Tuple[int, int]:
    """
    Function that gives the total shape of the embeddings array
    """

    tot_vec = 0
    vec_dim = -1

    for shape in read_shapes_local(embeddings_path):
        batch_length, dim = shape
        tot_vec += batch_length
        vec_dim = dim

    return tot_vec, vec_dim


def build_index(
    embeddings_path: str,
    index_key: str,
    metric_type: Union[str, int],
    nb_vectors: int,
    current_memory_available: str,
    use_gpu: bool = False,
):
    """
    Function that returns an index on the numpy arrays stored on disk in the embeddings_path path.
    """

    # Instanciate the index
    with Timeit(f"-> Instanciate the index {index_key}", indent=2):

        # Convert metric_type to faiss type
        metric_type = to_faiss_metric_type(metric_type)

        # Get information for one partition
        _, vec_dim = next(read_shapes_local(embeddings_path))

        # Instanciate the index
        index = index_factory(vec_dim, index_key, metric_type)

    # Extract training vectors
    with Timeit("-> Extract training vectors", indent=2):

        # Determine the number of vectors necessary to train the index
        train_size = get_optimal_train_size(nb_vectors, index_key, current_memory_available, vec_dim)
        print(
            f"Will use {train_size} vectors to train the index, "
            f"{cast_bytes_to_memory_string(train_size*vec_dim*4)} of memory"
        )

        # Extract training vectors
        train_vectors = next(read_embeddings_local(embeddings_path, batch_size=train_size, verbose=True))

    # Instanciate the index and train it
    # pylint: disable=no-member
    if use_gpu:
        # if this fails, it means that the GPU version was not comp.
        assert (
            faiss.StandardGpuResources
        ), "FAISS was not compiled with GPU support, or loading _swigfaiss_gpu.so failed"
        res = faiss.StandardGpuResources()
        dev_no = 0
        # transfer to GPU (may be partial).
        index = faiss.index_cpu_to_gpu(res, dev_no, index)

    with Timeit(
        f"-> Training the index with {train_vectors.shape[0]} vectors of dim {train_vectors.shape[1]}", indent=2
    ):
        index.train(train_vectors)

    del train_vectors

    # Add the vectors to the index.
    with Timeit("-> Adding the vectors to the index", indent=2):
        batch_size = get_optimal_batch_size(vec_dim, current_memory_available)
        for vec_batch in read_embeddings_local(embeddings_path, batch_size=batch_size, verbose=True):
            index.add(vec_batch)

    # Give standard values for index hyperparameters if possible.
    if any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
        set_search_hyperparameters(index, f"nprobe={64},efSearch={128},ht={2048}", use_gpu)

    # return the index.
    return index



""" Useful decorators for fast debuging """

import functools
import time
from contextlib import ContextDecorator
from datetime import datetime
from typing import Optional


class Timeit(ContextDecorator):
    """Timing class, used as a context manager"""

    def __init__(self, comment: Optional[str] = None, indent: int = 0, verbose: bool = True):
        self.start_time = 0
        self.comment = comment
        self.indent = indent
        self.verbose = verbose

    def __enter__(self):
        if self.verbose:
            if self.comment is not None:
                space = "\t" * self.indent
                start_date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                print(f"{space}{self.comment} {start_date}", flush=True)
            self.start_time = time.perf_counter()

    def __exit__(self, *exc):
        if self.verbose:
            end_time = time.perf_counter()
            run_time = end_time - self.start_time
            space = "\t" * self.indent
            print(f'{space}>>> Finished "{self.comment}" in {run_time:.4f} secs')


timeit = Timeit()


def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


def should_run_once(func):
    """
    Decorator to force a function to run only once.
    The fonction raises a ValueError otherwise.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if wrapper.has_run:
            raise ValueError("Can't call this function twice")
        wrapper.has_run = True
        return func(*args, **kwargs)

    wrapper.has_run = False
    return wrapper



""" functions that transform objects from the datasets """

import os
import re
from functools import partial
from multiprocessing import Pool
from typing import Optional

import numpy as np
import pyarrow.parquet as pq
from tqdm import tqdm as tq

from autofaiss.datasets.readers.remote_iterators import read_filenames
from autofaiss.utils.os_tools import list_local_files


def parquet_already_transformed(remote_embeddings_path: str, local_embeddings_path: str) -> bool:
    """Checks if embeddings were all transformed in .npy arrays or should be downloaded again"""

    remote_filenames = read_filenames(remote_embeddings_path)
    nb_remote_parquet_files = len(remote_filenames)

    local_filenames = list_local_files(local_embeddings_path)
    reg_exp_pattern = r"(emb|map)_\d+\.npy"
    reg_exp = re.compile(reg_exp_pattern)
    nb_local_map_and_emb_files = len([1 for filename in local_filenames if reg_exp.match(filename)])

    return 2 * nb_remote_parquet_files == nb_local_map_and_emb_files


def convert_parquet_to_numpy(
    parquet_file: str,
    embeddings_path: str,
    embedding_column_name: str,
    keys_path: Optional[str] = None,
    key_column_name: Optional[str] = None,
) -> None:
    """ Convert one embedding parquet file to an embedding numpy file """

    emb = None
    if not os.path.exists(embeddings_path):
        emb = pq.read_table(parquet_file).to_pandas()
        embeddings_raw = emb[embedding_column_name].to_numpy()
        embeddings = np.stack(embeddings_raw).astype("float32")
        np.save(embeddings_path, embeddings)

    if keys_path is not None and not os.path.exists(keys_path):
        if emb is None:
            emb = pq.read_table(parquet_file).to_pandas()
        key_raw = emb[key_column_name].to_numpy()
        np.save(keys_path, key_raw)


def run_one(
    parquet_file: str,
    embeddings_folder: str,
    delete: bool,
    embedding_column_name: str,
    keys_folder: Optional[str] = None,
    key_column_name: Optional[str] = None,
) -> None:
    """ Convertion function to call for parallel execution """
    num = parquet_file.split("/")[-1].split("-")[1]

    convert_parquet_to_numpy(
        parquet_file,
        f"{embeddings_folder}/emb_{num}.npy",
        embedding_column_name,
        f"{keys_folder}/key_{num}.npy",
        key_column_name,
    )

    if delete:
        os.remove(parquet_file)


def convert_all_parquet_to_numpy(
    parquet_folder: str,
    embeddings_folder: str,
    n_cores: int = 32,
    delete: bool = False,
    embedding_column_name: str = "embedding",
    keys_folder: Optional[str] = None,
    key_column_name: Optional[str] = None,
) -> None:
    """ Convert embedding parquet files to an embedding numpy files
    Optionally also extract keys from parquet files
    """

    assert n_cores > 0
    os.makedirs(embeddings_folder, exist_ok=True)
    if keys_folder:
        os.makedirs(keys_folder, exist_ok=True)

    parquet_files = [f"{parquet_folder}/{x}" for x in os.listdir(parquet_folder) if x.endswith(".parquet")]
    parquet_files.sort()

    nb_files = len(parquet_files)

    func = partial(
        run_one,
        embeddings_folder=embeddings_folder,
        delete=delete,
        embedding_column_name=embedding_column_name,
        keys_folder=keys_folder,
        key_column_name=key_column_name,
    )

    with tq(total=nb_files) as pbar:
        with Pool(processes=n_cores) as pool:
            for _ in pool.imap_unordered(func, parquet_files):
                pbar.update(1)



""" context manager to make reading/writing on hdfs as simple as writing on a local file """

from typing import Optional

import pyarrow
from pyarrow.filesystem import FileSystem


class HDFSFileSystem:
    """Context aware HDFSFileSystem using pyarrow.hdfs.
    Open and closes connection to HDFS thanks to a context manager
    >>> from deepr.io import HDFSFileSystem
    >>> with HDFSFileSystem() as fs:  # doctest: +SKIP
    ...     fs.open("path/to/file")  # doctest: +SKIP
    """

    def __init__(self):
        self._hdfs = None

    def __enter__(self):
        self._hdfs = pyarrow.hdfs.connect()
        return self._hdfs

    def __exit__(self, type, value, traceback):
        # pylint: disable=redefined-builtin
        self._hdfs.close()

    def __getattr__(self, name):
        # Expose self._hdfs methods and attributes (mimic inheritance)
        return getattr(self._hdfs, name)


class HDFSFile:
    """FileSystemFile, support of "r", "w" modes, readlines and iter.
    Makes it easier to read or write file from any filesystem. For
    example, if you use HDFS you can do
    >>> from deepr.io import HDFSFileSystem
    >>> with HDFSFileSystem() as fs:
    ...     with HDFSFile(fs, "viewfs://root/user/foo.txt", "w") as file:  # doctest: +SKIP
    ...         file.write("Hello world!")  # doctest: +SKIP
    The use of context manager means that the connection to the
    filesystem is automatically opened / closed, as well as the file
    buffer.
    Attributes
    ----------
    filesystem : FileSystem
        FileSystem instance
    path : str
        Path to file
    mode : str, Optional
        Write / read mode. Supported: "r", "rb" (default), "w", "wb".
    """

    def __init__(self, filesystem: FileSystem, path: str, mode: str = "rb", encoding: Optional[str] = "utf-8"):
        self.filesystem = filesystem
        self.path = path
        self.mode = mode
        self.encoding = None if "b" in mode else encoding
        self._file = filesystem.open(self.path, mode={"r": "rb", "w": "wb"}.get(mode, mode))

    def __iter__(self):
        yield from self.readlines()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # pylint: disable=redefined-builtin
        return self._file.__exit__(type, value, traceback)

    def __getattr__(self, name):
        # Expose self._file methods and attributes (mimic inheritance)
        return getattr(self._file, name)

    def write(self, data):
        if self.mode == "w":
            self._file.write(data.encode(encoding=self.encoding))
        elif self.mode == "wb":
            self._file.write(data)
        else:
            raise ValueError(f"Mode {self.mode} unkown (must be 'w' or 'wb').")

    def read(self):
        if self.mode == "r":
            return self._file.read().decode(encoding=self.encoding)
        elif self.mode == "rb":
            return self._file.read()
        else:
            raise ValueError(f"Mode {self.mode} unkown (must be 'r' or 'rb')")

    def readlines(self):
        return self.read().split("\n" if self.mode == "r" else b"\n")



"""
Function that downloads a parquet file containing embeddings on the
current machine in the right format (several .npy files)
"""
import logging
from autofaiss.datasets.readers.downloader import download
from typing import Optional

from autofaiss.external.build import get_estimated_download_time_infos
from autofaiss.datasets.transforming import convert_all_parquet_to_numpy
from autofaiss.utils.decorators import Timeit


def download_from_hdfs(
    embeddings_hdfs_path: str,
    local_save_path: str,
    n_cores: int = 10,
    delete_tmp_files: bool = True,
    bandwidth_gbytes_per_sec: Optional[float] = None,
    verbose: bool = True,
    embedding_column_name: str = "embedding",
    keys_folder: Optional[str] = None,
    key_column_name: Optional[str] = None,
) -> None:
    """
    Download a file containing embeddings in parquet format on hdfs and
    convert it to several .npy files
    Optionally save keys in another folder (useful to keep track of what is each embedding)
    """
    if bandwidth_gbytes_per_sec is not None:
        infos, _ = get_estimated_download_time_infos(embeddings_hdfs_path, bandwidth_gbytes_per_sec)
        print(infos)

    with Timeit("Download the parquet files on local disk", indent=1):
        download(embeddings_hdfs_path, dest_path=local_save_path, n_cores=n_cores, verbose=verbose)

    with Timeit("Convert .parquet files to numpy arrays", indent=1):
        convert_all_parquet_to_numpy(
            local_save_path,
            local_save_path,
            delete=delete_tmp_files,
            embedding_column_name=embedding_column_name,
            n_cores=n_cores,
            keys_folder=keys_folder,
            key_column_name=key_column_name,
        )


def main():
    """Main entry point"""
    logging.basicConfig(level=logging.INFO)
    fire.Fire(download_from_hdfs)


if __name__ == "__main__":
    main()



""" main file to create an index from the the begining """

import logging
import os
from pprint import pprint as pp
from typing import Dict, Optional, Union

import faiss

import fire
from autofaiss.external.build import (
    build_index,
    estimate_memory_required_for_index_creation,
    get_estimated_construction_time_infos,
    get_nb_vectors_and_dim,
)
from autofaiss.external.optimize import get_optimal_hyperparameters, get_optimal_index_keys_v2
from autofaiss.external.scores import compute_fast_metrics, compute_medium_metrics
from autofaiss.indices.index_utils import set_search_hyperparameters, load_index_from_hdfs
from autofaiss.utils.decorators import Timeit
from autofaiss.utils.cast import cast_memory_to_bytes, cast_bytes_to_memory_string


class Quantizer:
    """class defining the quantization pipeline"""

    def __init__(self):
        """Empty constructor"""

    def quantize(
        self,
        embeddings_path: str,
        output_path: str,
        index_key: Optional[str] = None,
        index_param: Optional[str] = None,
        max_index_query_time_ms: float = 10.0,
        max_index_memory_usage: str = "32G",
        current_memory_available: str = "32G",
        use_gpu: bool = False,
        metric_type: str = "ip",
    ) -> str:
        """
        Reads embeddings and creates a quantized index from them.
        The index is stored on the current machine at the given ouput path.

        Parameters
        ----------
        embeddings_path : str
            Local path containing all preprocessed vectors and cached files.
            Files will be added if empty.
        output_path: str
            Destination path of the quantized model on local machine.
        index_key: Optinal(str)
            Optional string to give to the index factory in order to create the index.
            If None, an index is chosen based on an heuristic.
        index_param: Optional(str)
            Optional string with hyperparameters to set to the index.
            If None, the hyper-parameters are chosen based on an heuristic.
        max_index_query_time_ms: float
            Bound on the query time for KNN search, this bound is approximative
        max_index_memory_usage: str
            Maximum size allowed for the index, this bound is strict
        current_memory_available: str
            Memory available on the machine creating the index, having more memory is a boost
            because it reduces the swipe between RAM and disk.
        use_gpu: bool
            Experimental, gpu training is faster, not tested so far
        metric_type: str
            Similarity function used for query:
                - "ip" for inner product
                - "l2" for euclidian distance
        """

        with Timeit("Launching the whole pipeline"):

            nb_vectors, vec_dim = get_nb_vectors_and_dim(embeddings_path)

            with Timeit("Compute estimated construction time of the index", indent=1):
                print(get_estimated_construction_time_infos(nb_vectors, vec_dim, indent=2))

            with Timeit("Checking that your have enough memory available to create the index", indent=1):
                necessary_mem, index_key_used = estimate_memory_required_for_index_creation(
                    nb_vectors, vec_dim, index_key, max_index_memory_usage
                )
                print(
                    f"{cast_bytes_to_memory_string(necessary_mem)} of memory "
                    "will be needed to train the index (more might be used if you have more)"
                )

                prefix = "(default) " if index_key is None else ""

                if 0.9 * necessary_mem > cast_memory_to_bytes(current_memory_available):

                    return (
                        f"The current memory available on your machine ({current_memory_available}) is not "
                        f"enough to create the {prefix}index {index_key_used} that requires "
                        f"{cast_bytes_to_memory_string(necessary_mem)} to train. "
                        "You can decrease the number of clusters of you index since the Kmeans algorithm "
                        "used for clusterisation is responsible for this high memory usage."
                    )

            if index_key is None:
                with Timeit("Selecting most promising index types given data characteristics", indent=1):
                    best_index_keys = get_optimal_index_keys_v2(nb_vectors, vec_dim, max_index_memory_usage)
                    if not best_index_keys:
                        return "Constraint on memory too high, no index can be that small"
                    index_key = best_index_keys[0]

            with Timeit("Creating the index", indent=1):
                index = build_index(
                    embeddings_path, index_key, metric_type, nb_vectors, current_memory_available, use_gpu
                )

            if index_param is None:
                with Timeit("Computing best hyperparameters", indent=1):
                    index_param = get_optimal_hyperparameters(index, index_key, max_speed=max_index_query_time_ms)

            # Set search hyperparameters for the index
            set_search_hyperparameters(index, index_param, use_gpu)
            print(f"The best hyperparameters are: {index_param}")

            # Save the index
            index_name = f"{index_key}-{index_param}.index"

            with Timeit("Saving the index on local disk", indent=1):
                os.makedirs(output_path, exist_ok=True)
                index_name = f"{index_key}-{index_param}.index"
                faiss.write_index(index, f"{output_path}/{index_name}")

            metric_infos: Dict[str, Union[str, float, int]] = {}

            with Timeit("Compute fast metrics", indent=1):
                metric_infos.update(compute_fast_metrics(embeddings_path, index))

            print("Recap:")
            pp(metric_infos)

        return "Done"

    def tuning(
        self,
        index_path: str,
        index_key: str,
        index_param: Optional[str] = None,
        dest_path: Optional[str] = None,
        is_local_index_path: bool = False,
        max_index_query_time_ms: float = 10.0,
        use_gpu: bool = False,
    ) -> str:
        """
        Set hyperparameters to the given index.

        If an index_param is given, set this hyperparameters to the index,
        otherwise perform a greedy heusistic to make the best out or the max_index_query_time_ms constraint

        Parameters
        ----------
        index_path : str
            Path to .index file on local disk if is_local_index_path is True,
            otherwise path on hdfs.
        index_key: str
            String to give to the index factory in order to create the index.
        index_param: Optional(str)
            Optional string with hyperparameters to set to the index.
            If None, the hyper-parameters are chosen based on an heuristic.
        dest_path: Optional[str]
            Path to the newly created .index file. On local disk if is_local_index_path is True,
            otherwise on hdfs. If None id given, index_path is the destination path.
        is_local_index_path: bool
            True if the dest_path and index_path are local path, False if there are hdfs paths.
        max_index_query_time_ms: float
            Query speed constraint for the index to create.
        use_gpu: bool
            Experimental, gpu training is faster, not tested so far.

        Returns
        -------
        infos: str
            Message describing the index created.
        """

        if dest_path is None:
            dest_path = index_path

        if is_local_index_path:
            with Timeit("Loading index from local disk"):
                index = faiss.read_index(index_path)
        else:
            with Timeit("Loading index from HDFS"):
                index = load_index_from_hdfs(index_path)

        if index_param is None:

            with Timeit("Compute best hyperparameters"):
                index_param = get_optimal_hyperparameters(index, index_key, max_speed=max_index_query_time_ms)

        with Timeit("Set search hyperparameters for the index"):
            set_search_hyperparameters(index, index_param, use_gpu)

        if is_local_index_path:
            with Timeit("Saving index on local disk"):
                faiss.write_index(index, dest_path)
        else:
            with Timeit("Saving index to HDFS"):
                load_index_from_hdfs(dest_path)

        return f"The optimal hyperparameters are {index_param}, the index with these parameters has been saved."

    def score_index(
        self,
        index_path: str,
        embeddings_path: str,
        is_local_index_path: bool = False,
        current_memory_available: str = "32G",
    ) -> None:
        """
        Compute metrics on a given index, use cached ground truth for fast scoring the next times.

        Parameters
        ----------
        index_path : str
            Path to .index file on local disk if is_local_index_path is True,
            otherwise path on hdfs.
        embeddings_path: str
            Local path containing all preprocessed vectors and cached files.
        is_local_index_path: bool
            True if the dest_path and index_path are local path, False if there are hdfs paths.
        current_memory_available: str
            Memory available on the current machine, having more memory is a boost
            because it reduces the swipe between RAM and disk.
        """

        if is_local_index_path:
            with Timeit("Loading index from local disk"):
                index = faiss.read_index(index_path)
        else:
            with Timeit("Loading index from HDFS"):
                index = load_index_from_hdfs(index_path)

        infos: Dict[str, Union[str, float, int]] = {}

        with Timeit("Compute fast metrics"):
            infos.update(compute_fast_metrics(embeddings_path, index))

        print("Intermediate recap:")
        pp(infos)

        with Timeit("Compute medium metrics"):
            infos.update(compute_medium_metrics(embeddings_path, index, current_memory_available))

        print("Performances recap:")
        pp(infos)


def main():
    """Main entry point"""
    logging.basicConfig(level=logging.INFO)

    fire.Fire(Quantizer)


if __name__ == "__main__":
    main()



""" Functions to find optimal index parameters """

import re
from functools import partial, reduce
from math import log2, sqrt, floor
from operator import mul
from typing import List, Optional

from autofaiss.indices.index_utils import set_search_hyperparameters, speed_test_ms_per_query
from autofaiss.utils.algorithms import discrete_binary_search
from autofaiss.utils.cast import cast_memory_to_bytes
from autofaiss.utils.decorators import Timeit


def get_optimal_train_size(
    nb_vectors: int, index_key: str, current_memory_available: Optional[str], vec_dim: Optional[int]
) -> int:
    """
    Function that determines the number of training points necessary to
    train the index, based on faiss heuristics for k-means clustering.
    """

    train_size = nb_vectors

    matching = re.findall(r"IVF\d+|IMI\d+x\d+", index_key)

    if matching:

        nb_clusters = nb_vectors
        # case IVF index
        if re.findall(r"IVF\d+", matching[0]):
            nb_clusters = int(matching[0][3:])
        # case IMI index
        elif re.findall(r"IMI\d+x\d+", matching[0]):
            nb_clusters = 2 ** reduce(mul, [int(num) for num in re.findall(r"\d+", matching[0])])

        points_per_cluster: float = 100

        # compute best possible number of vectors to give to train the index
        # given memory constraints
        if current_memory_available and vec_dim:
            size = cast_memory_to_bytes(current_memory_available)
            points_per_cluster = max(min(size / (4.0 * nb_clusters * vec_dim), points_per_cluster), 31.0)

        # You will need between 30 * nb_clusters and 256 * nb_clusters to train the index
        train_size = min(round(points_per_cluster * nb_clusters), nb_vectors)

    return train_size


def get_optimal_batch_size(vec_dim: int, current_memory_available: str) -> int:
    """compute optimal batch size to use the RAM at its full potential for adding vectors"""

    memory = cast_memory_to_bytes(current_memory_available)

    batch_size = min(
        2 ** 14, int(memory / (vec_dim * 4))
    )  # larger batch size than 16384 do not improve add speed and consume more ram

    return batch_size


def get_optimal_nb_clusters(nb_vectors: int) -> List[int]:
    """
    Returns a list with the recommended number of clusters for an index containing nb_vectors vectors.
    The first value is the most recommended one.
    see: https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index
    """

    nb_clusters_list = []

    if nb_vectors < 1_000_000:
        # no need to use HNSW for small number of clusters
        x_initial = 4 * sqrt(nb_vectors)  # between 4xsqrt(n) and 16xsqrt(n)
        x_closest_power = 2 ** round(log2(x_initial))
        nb_clusters_list.append(round(x_closest_power))
        nb_clusters_list.append(2 * x_closest_power)
        nb_clusters_list.append(x_closest_power)
        nb_clusters_list.append(round(x_initial))
    elif nb_vectors < 10_000_000:
        nb_clusters_list.append(16_384)
        nb_clusters_list.append(65_536)
    elif nb_vectors < 300_000_000:
        nb_clusters_list.append(65_536)
        nb_clusters_list.append(2 ** 17)
        nb_clusters_list.append(2 ** 18)  # slow training !
    else:
        nb_clusters_list.append(2 ** 17)
        nb_clusters_list.append(2 ** 18)  # slow training !
        nb_clusters_list.append(65_536)
        nb_clusters_list.append(2 ** 20)  # very slow training !

    nb_clusters_list = [int(x) for x in nb_clusters_list if x >= 256]

    if not nb_clusters_list:
        return [256]  # minimum possible value for Faiss IVF

    return nb_clusters_list


def get_optimal_index_keys_v2(
    nb_vectors: int,
    dim_vector: int,
    max_index_memory_usage: str,
    flat_threshold: int = 1000,
    hnsw_threshold: int = 10000,
    force_pq: Optional[int] = None,
):
    """
    Gives a list of interesting indices to try, *the one at the top is the most promising*

    See: https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index for
    detailed explanations.
    """

    # Exception cases:
    if nb_vectors < flat_threshold:  # HNSW Faiss slower when less than 1000 vectors
        return ["Flat"]
    if nb_vectors < hnsw_threshold:  # HNSW is faster and quantization is not possible (>=10_000 vectors needed)
        return ["HNSW15"]
    if force_pq is not None:
        return get_optimal_quantization(nb_vectors, dim_vector, force_quantization_value=force_pq)

    # Get max memory usage
    max_size_in_bytes = cast_memory_to_bytes(max_index_memory_usage)

    # If we can build an HNSW with the given memory constraints, it's the best
    m_hnsw = int(floor((max_size_in_bytes / (4 * nb_vectors) - dim_vector) / 2))
    if m_hnsw >= 8:
        return [f"HNSW{min(m_hnsw, 32)}"]

    # Otherwise, there is not enough space, let's go for quantization
    return get_optimal_quantization(nb_vectors, dim_vector, force_max_index_memory_usage=max_index_memory_usage)


def get_optimal_quantization(
    nb_vectors: int,
    dim_vector: int,
    force_quantization_value: Optional[int] = None,
    force_max_index_memory_usage: Optional[str] = None,
) -> List[str]:
    """
    Function that returns a list of relevant index_keys to create quantized indices.

    Parameters:
    ----------
    nb_vectors: int
        Number of vectors in the dataset.
    dim_vector: int
        Dimension of the vectors in the dataset.
    force_quantization_value: Optional[int]
        Force to use this value as the size of the quantized vectors (PQx).
        It can be used with the force_max_index_memory_usage parameter,
        but the result might be empty.
    force_max_index_memory_usage: Optional[str]
        Add a memory constraint on the index.
        It can be used with the force_quantization_value parameter,
        but the result might be empty.

    Return:
    -------
    index_keys: List[str]
        List of index_keys that would be good choices for quantization.
        The list can be empty if the given constraints are too strong.
    """

    # Default values
    pq_values = [64, 48, 32, 24, 16, 8, 4]
    targeted_compression_ratio = 0.0  # 0 = no constraint

    # Force compression ratio if required
    if force_max_index_memory_usage is not None:
        total_bytes = 4.0 * nb_vectors * dim_vector  # x4 because float32
        max_mem_bytes = float(cast_memory_to_bytes(force_max_index_memory_usage))
        targeted_compression_ratio = total_bytes / max_mem_bytes

    # Force quantization value if required
    if force_quantization_value is not None:
        pq_values = [force_quantization_value]

    # Compute optimal number of clusters
    relevant_list: List[str] = []
    nb_clusters_list = get_optimal_nb_clusters(nb_vectors)

    # Look for matching index keys
    for pq in pq_values:
        if pq < dim_vector:

            for nb_clusters in nb_clusters_list:

                # Compute quantized vector size
                cluster_size_byte = 1 + (log2(nb_clusters) - 1) // 8
                vector_size_byte = pq + cluster_size_byte

                # Compute compression ratio with quantization PQx
                compression_ratio = (4 * dim_vector) / vector_size_byte

                # Add index_key if compression ratio is high enough
                if compression_ratio >= targeted_compression_ratio:

                    # y is a multiple of pq (required)
                    # y <= d, with d the dimension of the input vectors (preferable)
                    # y <= 6*pq (preferable)
                    # here we choose a y slightly bigger than d to avoid losing information
                    # in case such as 101, 128 is better than 64 to avoid losing information
                    # in the linear transform
                    y = (min(dim_vector // pq, 6) + 1) * pq
                    cluster_opt = f"IVF{nb_clusters}" if nb_clusters < 1000 else f"IVF{nb_clusters}_HNSW32"
                    relevant_list.append(f"OPQ{pq}_{y},{cluster_opt},PQ{pq}x8")

    return relevant_list


def get_optimal_hyperparameters(
    index,
    index_key: str,
    max_speed: float,  # milliseconds
    use_gpu: bool = False,
    max_timeout_per_iteration: float = 1.0,  # seconds
    min_ef_search: int = 32,
) -> str:
    """Find the optimal hyperparameters to maximize the recall given a query speed in milliseconds/query"""

    # nb_vectors = index.ntotal

    if any(re.findall(r"OPQ\d+_\d+,IVF\d+,PQ\d+", index_key)):
        with Timeit("starting initial reconstruct"):
            query = index.reconstruct_n(0, 4000)
        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 15 * max_speed / 1000),
        )

        #
        nprobes = list(range(max(1, min_ef_search // 2), 6144))
        nprobes.sort()

        ht = 2048

        def is_not_acceptable_speed(rank_nprobe):

            nprobe = nprobes[rank_nprobe]
            param_str = f"nprobe={nprobe},ht={ht}"
            set_search_hyperparameters(index, param_str, use_gpu)
            speed = get_speed(index)

            return speed >= max_speed

        with Timeit("starting initial binary search"):
            rank_best_nprobe = discrete_binary_search(is_not_acceptable_speed, len(nprobes)) - 1

        # make sure that the query time is respected
        with Timeit("starting final reconstruct"):
            query = index.reconstruct_n(0, min(index.ntotal, 50000))

        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 200 * max_speed / 1000),
        )

        with Timeit("final linear search"):
            while is_not_acceptable_speed(rank_best_nprobe) and rank_best_nprobe > 1:
                rank_best_nprobe -= 1

        rank_best_nprobe = max(0, min(rank_best_nprobe, len(nprobes) - 1))

        nprobe = nprobes[rank_best_nprobe]

        param_str = f"nprobe={nprobe},ht={ht}"

        return param_str

    elif any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):

        with Timeit("starting initial reconstruct"):
            query = index.reconstruct_n(0, 4000)
        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 15 * max_speed / 1000),
        )

        #
        nprobes = list(range(max(1, min_ef_search // 2), 6144))
        nprobes.sort()

        ht = 2048

        def is_not_acceptable_speed(rank_nprobe):

            nprobe = nprobes[rank_nprobe]
            param_str = f"nprobe={nprobe},efSearch={2*nprobe},ht={ht}"
            set_search_hyperparameters(index, param_str, use_gpu)
            speed = get_speed(index)

            return speed >= max_speed

        with Timeit("starting initial binary search"):
            rank_best_nprobe = discrete_binary_search(is_not_acceptable_speed, len(nprobes)) - 1

        # make sure that the query time is respected
        with Timeit("starting final reconstruct"):
            query = index.reconstruct_n(0, min(index.ntotal, 50000))

        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 200 * max_speed / 1000),
        )

        with Timeit("final linear search"):
            while is_not_acceptable_speed(rank_best_nprobe) and rank_best_nprobe > 1:
                rank_best_nprobe -= 1

        rank_best_nprobe = max(0, min(rank_best_nprobe, len(nprobes) - 1))

        nprobe = nprobes[rank_best_nprobe]

        param_str = f"nprobe={nprobe},efSearch={2*nprobe},ht={ht}"

        return param_str

    if any(re.findall(r"HNSW\d+", index_key)):

        query = index.reconstruct_n(0, min(index.ntotal, 4000))
        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 15 * max_speed / 1000),
        )

        ef_searchs = list(range(16, 2 ** 14))

        def is_not_acceptable_speed_ef_search(rank_efsearch):

            ef_search = ef_searchs[rank_efsearch]
            param_str = f"efSearch={ef_search}"
            set_search_hyperparameters(index, param_str, use_gpu)
            speed = get_speed(index)

            return speed >= max_speed

        rank_best_ef_search = discrete_binary_search(is_not_acceptable_speed_ef_search, len(ef_searchs)) - 1

        # make sure that the query time is respected
        query = index.reconstruct_n(0, min(index.ntotal, 50000))
        get_speed = partial(
            speed_test_ms_per_query,
            query=query,
            ksearch=40,
            timout_s=min(max_timeout_per_iteration, 100 * max_speed / 1000),
        )

        while is_not_acceptable_speed_ef_search(rank_best_ef_search) and rank_best_ef_search > 1:
            rank_best_ef_search -= 1

        rank_best_ef_search = max(0, min(rank_best_ef_search, len(ef_searchs) - 1))

        ef_search = ef_searchs[rank_best_ef_search]

        param_str = f"efSearch={ef_search}"

        return param_str

    if index_key == "Flat":
        return ""

    raise NotImplementedError("we don't have heuristics for that kind or index")



import numpy as np
import random
from autofaiss.datasets.readers.local_iterators import read_embeddings_local
import shutil
import os


def build_test_collection(min_size=2, max_size=10000, dim=512, nb_files=5):
    tmp_dir = "/tmp/autofaiss_test"
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.mkdir(tmp_dir)
    sizes = [random.randint(min_size, max_size) for _ in range(nb_files)]
    dim = dim
    for i, size in enumerate(sizes):
        arr = np.random.rand(size, dim).astype("float32")
        np.save(str(tmp_dir) + "/" + str(i) + ".npy", arr)
    return tmp_dir, sizes, dim


def test_read_embeddings_local():
    tmp_dir, sizes, dim = build_test_collection(min_size=2, max_size=12547, dim=512, nb_files=37)
    filenames = sorted([str(tmp_dir) + "/" + str(i) + ".npy" for i in range(len(sizes))])
    ref = np.concatenate([np.load(f) for f in filenames])
    batch_size = 62
    it = read_embeddings_local(tmp_dir, batch_size=batch_size)
    total_expected = sum(sizes)
    last = total_expected // batch_size
    last_expected_batch = total_expected - last * batch_size
    assert ref.shape[0] == total_expected

    total_found = 0
    current = 0
    for i, batch in enumerate(it):
        if last == i:
            expected_batch_size = last_expected_batch
        else:
            expected_batch_size = batch_size
        assert batch.shape[0] == expected_batch_size
        assert batch.shape[1] == dim
        np.testing.assert_array_equal(ref[current : (current + expected_batch_size)], batch)
        current += expected_batch_size
        total_found += batch.shape[0]
    assert total_expected == total_found



"""
Index metadata for Faiss indices.
"""

import re
from enum import Enum
from math import ceil, log2

from autofaiss.utils.cast import cast_bytes_to_memory_string
from autofaiss.external.descriptions import (
    INDEX_DESCRIPTION_BLOCKS,
    IndexBlock,
    TUNABLE_PARAMETERS_DESCRIPTION_BLOCKS,
    TunableParam,
)


class IndexType(Enum):
    FLAT = 0
    HNSW = 1
    OPQ_IVF_PQ = 2
    OPQ_IVF_HNSW_PQ = 3
    PAD_IVF_HNSW_PQ = 4
    NOT_SUPPORTED = 5


class IndexMetadata:
    """
    Class to compute index metadata given the index_key, the number of vectors and their dimension.

    Note: We don't create classes for each index type in order to keep the code simple.
    """

    def __init__(self, index_key: str, nb_vectors: int, dim_vector: int):

        self.index_key = index_key
        self.nb_vectors = nb_vectors
        self.dim_vector = dim_vector
        self.fast_description = ""
        self.description_blocs = []
        self.tunable_params = []
        self.params = {}

        params = [int(x) for x in re.findall(r"\d+", index_key)]

        if any(re.findall(r"OPQ\d+_\d+,IVF\d+,PQ\d+", index_key)):
            self.index_type = IndexType.OPQ_IVF_PQ
            self.fast_description = "An inverted file index (IVF) with quantization and OPQ preprocessing."
            self.description_blocs = [IndexBlock.IVF, IndexBlock.PQ, IndexBlock.OPQ]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.HT]

            self.params["pq"] = params[3]
            self.params["nbits"] = params[4] if len(params) == 5 else 8  # default value
            self.params["ncentroids"] = params[2]
            self.params["out_d"] = params[1]
            self.params["M_OPQ"] = params[0]

        elif any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
            self.index_type = IndexType.OPQ_IVF_HNSW_PQ
            self.fast_description = "An inverted file index (IVF) with quantization, OPQ preprocessing, and HNSW index."
            self.description_blocs = [IndexBlock.IVF_HNSW, IndexBlock.HNSW, IndexBlock.PQ, IndexBlock.OPQ]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.EFSEARCH, TunableParam.HT]

            self.params["M_HNSW"] = params[3]
            self.params["pq"] = params[4]
            self.params["nbits"] = params[5] if len(params) == 6 else 8  # default value
            self.params["ncentroids"] = params[2]
            self.params["out_d"] = params[1]
            self.params["M_OPQ"] = params[0]

        elif any(re.findall(r"Pad\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
            self.index_type = IndexType.PAD_IVF_HNSW_PQ
            self.fast_description = (
                "An inverted file index (IVF) with quantization, a padding on input vectors, and HNSW index."
            )
            self.description_blocs = [IndexBlock.IVF_HNSW, IndexBlock.HNSW, IndexBlock.PQ, IndexBlock.PAD]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.EFSEARCH, TunableParam.HT]

            self.params["out_d"] = params[0]
            self.params["M_HNSW"] = params[2]
            self.params["pq"] = params[3]
            self.params["nbits"] = params[4] if len(params) == 5 else 8  # default value
            self.params["ncentroids"] = params[1]

        elif any(re.findall(r"HNSW\d+", index_key)):
            self.index_type = IndexType.HNSW
            self.fast_description = "An HNSW index."
            self.description_blocs = [IndexBlock.HNSW]
            self.tunable_params = [TunableParam.EFSEARCH]

            self.params["M_HNSW"] = params[0]

        elif index_key == "Flat":
            self.index_type = IndexType.FLAT
            self.fast_description = "A simple flat index."
            self.description_blocs = [IndexBlock.FLAT]
            self.tunable_params = []

        else:
            self.index_type = IndexType.NOT_SUPPORTED
            self.fast_description = "No description for this index, feel free to contribute :)"
            self.description_blocs = []
            self.tunable_params = []

    def get_index_type(self) -> IndexType:
        """
        return the index type.
        """
        return self.index_type

    def estimated_index_size_in_bytes(self) -> int:
        """
        Compute the estimated size of the index in bytes.
        """

        if self.index_type == IndexType.FLAT:
            return self.nb_vectors * self.dim_vector * 4

        if self.index_type == IndexType.HNSW:
            # M bidirectional links per vector in the HNSW graph
            hnsw_graph_in_bytes = self.nb_vectors * self.params["M_HNSW"] * 2 * 4
            vectors_size_in_bytes = self.nb_vectors * self.dim_vector * 4
            return vectors_size_in_bytes + hnsw_graph_in_bytes

        if self.index_type in [IndexType.OPQ_IVF_PQ, IndexType.OPQ_IVF_HNSW_PQ, IndexType.PAD_IVF_HNSW_PQ]:
            # We neglict the size of the OPQ table for the moment.
            code_size = ceil(self.params["pq"] * self.params["nbits"] / 8)
            cluster_size_byte = 1 + int((log2(self.params["ncentroids"]) - 1) // 8)
            vector_size_byte = code_size + cluster_size_byte

            vectors_size_in_bytes = self.nb_vectors * vector_size_byte
            centroid_size_in_bytes = self.params["ncentroids"] * self.dim_vector * 4

            total_size_in_byte = vectors_size_in_bytes + centroid_size_in_bytes

            if self.index_type in [IndexType.OPQ_IVF_HNSW_PQ, IndexType.PAD_IVF_HNSW_PQ]:
                total_size_in_byte += self.params["ncentroids"] * self.params["M_HNSW"] * 2 * 4

            if self.index_type in [IndexType.OPQ_IVF_PQ, IndexType.OPQ_IVF_HNSW_PQ]:
                total_size_in_byte += self.params["M_OPQ"] * self.params["out_d"] * 4

            return total_size_in_byte

        return -1

    def get_index_description(self, tunable_parameters_infos=False) -> str:
        """
        Gives a generic description of the index.
        """

        description = self.fast_description

        if self.index_type == IndexType.NOT_SUPPORTED:
            return description

        description += "\n"
        index_size_string = cast_bytes_to_memory_string(self.estimated_index_size_in_bytes())
        description += f"The size of the index should be around {index_size_string}.\n\n"
        description += "\n".join(INDEX_DESCRIPTION_BLOCKS[desc] for desc in self.description_blocs) + "\n\n"

        if tunable_parameters_infos:
            if not self.tunable_params:
                description += "No parameters can be tuned to find a query speed VS recall tradeoff\n\n"
            else:
                description += "List of parameters that can be tuned to find a query speed VS recall tradeoff:\n"
                description += (
                    "\n".join(TUNABLE_PARAMETERS_DESCRIPTION_BLOCKS[desc] for desc in self.tunable_params) + "\n\n"
                )

        description += """
For all indices except the flat index, the query speed can be adjusted.
The lower the speed limit the lower the recall. With a looser constraint
on the query time, the recall can be higher, but it is limited by the index
structure (if there is quantization for instance).
"""
        return description



# pylint: disable=unused-import,missing-docstring

import autofaiss.datasets
import autofaiss.external
import autofaiss.indices
import autofaiss.metrics
import autofaiss.utils
from autofaiss.version import __author__, __version__



""" Functions to download and transform the parquet files to numpy files """

import os
from itertools import repeat
from multiprocessing import Pool
from typing import Tuple

from tqdm import tqdm as tq

from autofaiss.datasets.readers.remote_iterators import read_filenames
from autofaiss.utils.os_tools import run_command


def download(parquet_embeddings_path: str, dest_path: str, n_cores: int = 32, verbose: bool = True) -> bool:
    """
    Download .parquet files from hdfs at max speed
    Parallelisation is essential to use the full bandwidth.
    """

    filenames = read_filenames(parquet_embeddings_path)

    nb_files = len(filenames)

    os.makedirs(dest_path, exist_ok=True)

    src_dest_paths = zip(filenames, repeat(dest_path))

    if n_cores == 1:

        if verbose:
            src_dest_paths = tq(list(src_dest_paths))

        for src_dest_path in src_dest_paths:
            download_one(src_dest_path)

    else:

        with tq(total=nb_files) as pbar:
            with Pool(processes=n_cores) as pool:
                for _ in pool.imap_unordered(download_one, src_dest_paths):
                    pbar.update(1)

    return True


def download_one(src_dest_path: Tuple[str, str]) -> None:
    """Function to download one file from hdfs to local"""

    filename, dest_path = src_dest_path

    if not os.path.exists(f"{dest_path}/{filename.split('/')[-1]}"):
        cmd = f"hdfs dfs -copyToLocal {filename} {dest_path}"
        _ = run_command(cmd)
